%%%%
% Machine learning system for experimental purposes
%
% Author: Tony Lindgren
%
% version 0.0.3
%



:- module(rule_inducer,[run_experiment/3, start/0]).
:- use_module([library(lists), library(file_systems), library(random), library(codesio), tree, util, model_based_sampling, rnd_uniform, munge]).

:- volatile rule_no/1.
:- dynamic rule_no/1.
:- volatile numerical/1.
:- dynamic numerical/1.

rule_no(_).
put_rule_no:-
    retract(rule_no(T)),
    NewT is T + 1,
    assert(rule_no(NewT)).

setup:-
    current_directory(_,'your_path_here/code/data_f').

experiment([validation_size(0.0), fold_x_v(10)]).

method([tree([classification, transform(no), rnd_feature(no), min_cov(5), min_margin(0.90)]),
        ensemble(100, [classification, bagging, generate_ex([no, rnd_uniform(100),  rnd_uniform(250),
	rnd_uniform(500),  rnd_uniform(750), rnd_uniform(1000),  rnd_uniform(2500), rnd_uniform(5000),
	rnd_uniform(7500), rnd_uniform(10000),
        mbs(100), mbs(250), mbs(500), mbs(750), mbs(1000), mbs(2500), mbs(5000), mbs(7500), mbs(10000),
        munge(100, 25, 4), munge(250, 25, 4), munge(500, 25, 4), munge(750, 25, 4), munge(1000, 25, 4),
	munge(2500, 25, 4), munge(5000, 25, 4), munge(7500, 25, 4), munge(10000, 25, 4)
	]),
        rnd_feature(no), min_cov(5), min_margin(0.90)], dac)]).


data([%set(breast_cancer_wisconsin, relation(_,_,_,_,_,_,_,_,_,_), breast_cancer_wisconsin_cx, breast_cancer_wisconsin_ex_mod),
      %set(bupa, bupa(_,_,_,_,_,_,_), bupa, bupa_ex),
      %set(cleveland_heart_disease, relation(_,_,_,_,_,_,_,_,_,_,_,_,_,_), cleveland_heart_disease_cx, cleveland_heart_disease_ex_mod),
      %set(glass, glass(_,_,_,_,_,_,_,_,_,_), glass,glass_ex),
      %set(haberman, haberman(_,_,_,_), haberman, haberman_data),
      set(iris, iris(_,_,_,_,_), iris, iris_data)
      %set(thyroid, thyroid(_,_,_,_,_,_), new_thyroid, new_thyroid_ex),
      %set(wine, relation(_,_,_,_,_,_,_,_,_,_,_,_,_,_), wine_cx, wine_ex),
      %set(image_segmentation, relation(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), image_segmentation_cx, image_segmentation_ex)
      %set(ionosphere, good_radar(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), ionosphere, ionosphere_ex)
      %set(pendigitis, rel(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), pendigits_names, pendigits_ex),
      %set(pima_indians, diabetes(_,_,_,_,_,_,_,_,_), pima_indians, pima_indians_ex),   
      %set(waveform_noise2, rel(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), waveform_noise2_cx, waveform_noise2_ex
      %set(pop_failures, rel(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), pop_failures_cx, pop_failures_ex)          
      ]).


start:-
    experiment(E),
    method(M),
    data(D),
    write('Starting experiment'),nl,
    run_experiment(experiment(E), method(M), data(D)).

run_experiment(experiment(Parameters), method(Methods), data(DataSets)):-
    process_data(DataSets, Parameters, Methods).

process_data([], _, _).
process_data([set(Name, Head, BackgroundKnowledge, Examples)|Sets], Parameters, Methods):-
    read_from_file(BackgroundKnowledge, BKList),
    read_from_file(Examples, ExList),
    random_permutation(ExList, RndExL),
    member(fold_x_v(Folds), Parameters),
    member(validation_size(VSize),Parameters),
    pre_process_data(Folds, VSize, RndExL, ExLists),
    use_methods(Methods, Folds, Name, Head, Parameters, ExLists, BKList),!,
    process_data(Sets, Parameters, Methods).

use_methods([], _, _, _, _, _, _).
use_methods([Method|Methods], Folds, Name, Head, Parameters, ExLists, BKList):-
    do_experiment(Folds, Name, Head, Method, ExLists, BKList, Results),
    open('Result', append, Stream),
    write_results(Results, Stream, Folds),
    close(Stream),
    retract_loop(Results, Head, BKList),!,
    use_methods(Methods, Folds, Name, Head, Parameters, ExLists, BKList).

write_results([], _, _).
write_results([result(DataSet, Method, MethodSize, _Key, ClassT, NoRules, Eval, EvalTime)|Rest], Stream, Folds):-
    find_and_remove(Rest, DataSet, Method, MethodSize, NoRulesL, EvalList, EvalTimeList, NewRest),
    (is_list(NoRules)->
      sumlist(NoRules, TempRules),
      sum_list_c(NoRulesL, TempRules, TotRules)
    ;
      sum_list_c(NoRulesL, NoRules, TotRules)
    ),
    AverageRules is TotRules / Folds,
    summarize_evaluation(EvalList, Eval, EvalTimeList, EvalTime, AggregatedValues, AggregatedTime),
    nl(Stream),nl(Stream),
    nl(Stream),
    write(Stream,'================================'),nl(Stream),
    write(Stream,'Data set: '),
    write(Stream,DataSet),nl(Stream),
    write(Stream,'Method: '),
    write(Stream,Method),nl(Stream),
    write(Stream,'Size: '),
    write(Stream,MethodSize),nl(Stream),
    write(Stream,'--------------------------------'),nl(Stream),
    write(Stream,'Average results over all folds :'),nl(Stream),
    write(Stream,'--------------------------------'),nl(Stream),
    write(Stream,'Number of rules: '),
    write(Stream,AverageRules),nl(Stream),
    get_statistics(AggregatedValues, Stream, ClassT, _Accuracy, _AUC, _ROC, _Precision, _Recall, _ConfusionMatrix),
    write(Stream,'Time to classify test set: '),
    write(Stream,AggregatedTime),nl(Stream),
    write_results(NewRest, Stream, Folds).

find_and_remove([], _, _, _, [], [], [], []).
find_and_remove([result(Name, Method, MethodSize, _Key, _ClassT, NoRules, Eval, EvalTime)|Rest], Name, Method, MethodSize, [NoRules|NoRulesL], [Eval|EvalList], [EvalTime|EvalTimeList], NewRest):-
    !, find_and_remove(Rest, Name, Method, MethodSize, NoRulesL, EvalList, EvalTimeList, NewRest).
find_and_remove([Save|Rest], Name, Method, MethodSize, NoRulesL, EvalList, EvalTimeList, [Save|NewRest]):-
    find_and_remove(Rest, Name, Method, MethodSize, NoRulesL, EvalList, EvalTimeList, NewRest).

sum_list_c([], FVal, FVal).
sum_list_c([L|Ls], TVal, FVal):-
    number(L),
    NewTVal is TVal + L,!,
    sum_list_c(Ls, NewTVal, FVal).
sum_list_c([L|Ls], TVal, FVal):-
    is_list(L),
    sumlist(L,LVal),
    NewTVal is TVal + LVal,!,
    sum_list_c(Ls, NewTVal, FVal).

%Loops many times, not so efficent...
get_statistics(AggValues, Stream, _ClassTuples, Accuracy, _AUC, _ROC, _Precision, _Recall, _ConfusionMatrix):-
    get_accuracy(AggValues, 0, 0, Accuracy),
    write(Stream,'Accuracy: '),
    write(Stream, Accuracy),nl(Stream). %We only do accuracy
    %get_precision(AggValues, ClassTuples, Precision),
    %get_recall(AggValues, ClassTuples, Recall),
    %write(Stream,'Confusion matrix: '),nl,
    %format(Stream,'~w ~t ~w ~n',['Corr. class','Pred. class']),
    %length(ClassTuples, CTL),
    %write_f_row(ClassTuples, Stream,1),
    %get_confusion_mat(AggValues, ClassTuples, 0, CTL,  ConfusionMatrix).

write_f_row([], _, _):-
    nl.
write_f_row([_-Class|CT], Stream, I):-
    %Tab is I * 5,
    format(Stream,'~t ~t ~t ~t ~w',[Class]),
    write_f_row(CT, Stream,I).

get_confusion_mat(_, _, CTL, CTL, []).
get_confusion_mat(AggVals, ClassTuple, I, CTL, [Class-Row|Rec]):-
    NewI is I + 1,
    nth1(NewI, ClassTuple, _-Class),
    write_row(AggVals, Class, NewI, ClassTuple, Row), nl,
    get_confusion_mat(AggVals, ClassTuple, NewI, CTL, Rec).

write_row(AggVals, Class, I, CT, Row):-
    member(Class-[_, TP, _, _, Row], AggVals),
    write(Class),
    write_before(I, CT, Row),
    format('~t ~d ',[TP]),
    length(CT, ClassL),
    ClassL1 is ClassL + 1,
    NextI is I + 1,
    write_after(NextI, CT, ClassL1, Row).

write_after(ClassL, _, ClassL, _).
write_after(I, CT, ClassL, Row):-
    nth1(I, CT, C),
    write_rr(C, Row),
    NewI is I + 1,
    write_after(NewI, CT, ClassL, Row).

write_before(1, _, _).
write_before(I, [CT|CTs], Row):-
    write_rr(CT, Row),
    NewI is I - 1,
    write_before(NewI, CTs, Row).

%write_rr([],_):-
%    nl.
write_rr(_-Class,RR):-
    member(No-Class,RR),
    format('~t ~d ',[No]).%,
    %write_rr(CTs,RR).

get_recall([],[],[]).
get_recall(AggVals,[_-Class|CT],[Class-Recall|Rec]):-
    select(Class-[_, TP, _, FN, _],AggVals,Rest),
    Recall is TP / (TP + FN),
    write('Recall for class: '),write(Class),write(' is: '),write(Recall),nl,
    get_recall(Rest,CT,Rec).

get_precision([],[],[]).
get_precision(AggVals,[_-Class|CT],[Class-Precision|Prec]):-
    select(Class-[_, TP, FP, _, _],AggVals,Rest),
    Denominator is TP + FP,
    (Denominator = 0 ->
       Precision = 0
    ;
       Precision is TP / Denominator
    ),
    write('Precision for class: '),write(Class),write(' is: '),write(Precision),nl,
    get_precision(Rest,CT,Prec).

get_accuracy([], Tot, TP, Accuracy):-
    (Tot = 0 ->
      Accuracy = 0
    ;
      Accuracy is TP / Tot
    ).
get_accuracy([_-[No, TP, _, _, _]|R], TTot, TTP, Accuracy):-
    NewTot is TTot + No,
    NewTP is TTP + TP,
    get_accuracy(R, NewTot, NewTP, Accuracy).

retract_loop([], _, _).
retract_loop([result(_, _, _, Key, _, _, _, _)|Keys], Head, BKList):-
    retract_inner_loop(Key, Head, BKList),
    retract_loop(Keys, Head, BKList).

retract_inner_loop(Key, Head, BKList):-
    atom(Key),
    retractall(Key:Head),
    retract_all(BKList, Key).
retract_inner_loop([], _, _).
retract_inner_loop([Key|Keys], Head, BKList):-
    retractall(Key:Head),
    retract_all(BKList, Key),
    retract_inner_loop(Keys, Head, BKList).
retract_inner_loop(_, _, _). %Fail silently

summarize_evaluation([], FSum, [], TimeSum, FSum, TimeSum).
summarize_evaluation([Fold|Folds], Cov_TC_FC_FN_F, [FTime|FTimes], Time, FinalResult, FinalTime):-
    summarize_evaluation1(Fold, Cov_TC_FC_FN_F, FSum),
    NewTime is Time + FTime,
    summarize_evaluation(Folds, FSum, FTimes, NewTime, FinalResult, FinalTime).

summarize_evaluation1([], FinalSum, FinalSum).
summarize_evaluation1([Class-[Cov,TP,FP,FN,False]|Rest], TempCov_TP_FP_FN_F, FSum):-
    select(Class-[TempCov,TempTP,TempFP,TempFN, TempF], TempCov_TP_FP_FN_F, RestTempCov_TP_FP_FN_F),
    NewTempCov is TempCov + Cov,
    NewTempTP is TempTP + TP,
    NewTempFP is TempFP + FP,
    NewTempFN is TempFN + FN,
    sum_fn(False, TempF, NewTempF),
    append([Class-[NewTempCov, NewTempTP, NewTempFP, NewTempFN, NewTempF]], RestTempCov_TP_FP_FN_F, NewTempCov_TP_FP_FN_F),
    summarize_evaluation1(Rest,  NewTempCov_TP_FP_FN_F, FSum).

get_mean_val([], _Folds, _).
get_mean_val([Class-[ClassCov,TP,FP,FN]|Zz], Folds, MeanFSum):-
    MeanCov is ClassCov / Folds,
    MeanTP is TP / Folds,
    MeanFP is FP / Folds,
    write('Class: '), write(Class),nl,
    write('Covers '), write(MeanCov), write(' of the example space'),nl,
    write('Accuracy: '), write(MeanTP),nl,
    write(': '), write(MeanFP), nl,
    get_mean_fn(FN, Folds),
    get_mean_val(Zz, Folds, MeanFSum).

get_mean_fn(Cov-FN, Folds):-
    MeanCov is Cov / Folds,
    write('% wrongly classified not belonging to the particular class classified as such are: '), write(MeanCov),nl,
    get_mean_fn2(FN, Folds).

get_mean_fn2([], _Folds).
get_mean_fn2([Class-AggNo|FNs], Folds):-
    Percent is AggNo / Folds,
    write('Class '), write(Class), write(' wrongly classified in (%, wrt. all classes): '), write(Percent),nl,
    get_mean_fn2(FNs, Folds).

sum_fn([], _, []):-!.
sum_fn([No-Class|Rest], TempFN, [NewNo-Class|Zz]):-
    member(TempNo-Class, TempFN),
    NewNo is No + TempNo,
    sum_fn(Rest, TempFN, Zz).

get_parts([],_,[]).
get_parts([Class-No|Yz],TempFN,[Class-NewNo|Zz]):-
    member(Class-TempNo, TempFN),
    NewNo is No + TempNo,
    get_parts(Yz,TempFN,Zz).

get_tc_fc(Class-[TC,FC], TempTCFC, Class-[NewTempTC,NewTempFC]):-
    member(Class-[TempTC,TempFC],TempTCFC),
    NewTempTC is TC + TempTC,
    NewTempFC is FC + TempFC.

do_experiment(0, _, _, _, _, _, []).
do_experiment(Folds, Name, Head, Parameters, ExLists, BKList, Eval):-
    do_one_fold(Folds, Name, Head, Parameters, ExLists, BKList, PartEval),
    NewFolds is Folds - 1,!,
    do_experiment(NewFolds, Name, Head, Parameters, ExLists, BKList, RestEval),
    append(PartEval, RestEval, Eval).
do_experiment(Folds, Name, Head, Parameters, ExLists, BKList, Eval):-
  write(Folds), write(Name), write(Head), write(Parameters), write(ExLists), write(BKList), write(Eval),
  do_experiment(Folds, Name, Head, Parameters, ExLists, BKList, Eval).

do_one_fold(TestFold, Name, Head, Method, ExLists, BKList, Result):-
    TestFold \== 0,
    nth1(TestFold, ExLists, TestExList, TrainExList),
    get_parameters(Method, Parameters),
    do_induction(Method, Head, TrainExList, BKList, TestFold, ClassIndex, Keys, _Trees, NoRules, Rules),
    Keys = [Key|_],
    create_class_tuples(Key, ClassTuples),
    statistics(runtime, _),
    classify_test_set(TestExList, ClassIndex, ClassTuples, ClassTuples, ClassDist, Keys, Res),
    check_ties(Res, NoTies),
    write('We found this amount of ties: ' ), write(NoTies),nl,
    statistics(runtime, [_,EnsEvalTime]),
    length(Res, ResL),
    evaluate(ClassDist, ClassTuples, Res, ResL, PartEval),!,
    (ensemble(Size, Settings, _) = Method ->
      member(generate_ex(List), Settings),
      copy_term(ClassTuples, CClassTuples),
      join_folds(TrainExList, NewTrainExList),
      use_sampling_methods(List, Rules, Name, NewTrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, CClassTuples, SampRes),
      append([result(Name, ensemble, Size, Keys, ClassTuples, NoRules, PartEval, EnsEvalTime)], SampRes, Result)
    ;
      Result = [result(Name, tree, 1, Keys, ClassTuples, NoRules, PartEval, EnsEvalTime)]
    ).

%%
% Check parameter settings to call correct sampling method, then use it for ruleinduction and evaluate the result...
%
use_sampling_methods([], _, _, _, _, _, _, _, _, _, _, []).
use_sampling_methods([no], _, _, _, _, _, _, _, _, _, _, []).
use_sampling_methods([no|R], RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, CClassTuples, SampRes):-
    !,use_sampling_methods(R, RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, CClassTuples, SampRes).
use_sampling_methods([rnd_uniform(SampleSize)|R], RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, [PartResult|RestRes]):-
    write('starting rnd_uniform with size: '), write(SampleSize),nl,
    write_to_codes(rnd_uniform-TestFold-SampleSize, TreeCodes),
    atom_codes(TreeKey,TreeCodes),
    rnd_uniform(TrainExList, RulesL, SampleSize, Head, ClassIndex, ClassT, GeneratedEx),
    append(TrainExList, GeneratedEx, CMMTrainEx),
    tree(CMMTrainEx, BKList, Parameters, Head, TreeKey, ClassIndex, Tree, _DefaultFact),
    evaluate_tree(TestExList, Name, rnd_uniform, SampleSize, Tree, BKList, Head, TreeKey, ClassIndex, ClassT, PartResult),!,
    use_sampling_methods(R, RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, RestRes).
use_sampling_methods([munge(SampleSize, Prob, Variance)|R], RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, [PartResult|RestRes]):-
    write('starting munge with size: '), write(SampleSize),nl,
    write_to_codes(munge-TestFold-SampleSize, TreeCodes),
    atom_codes(TreeKey,TreeCodes),
    munge(TrainExList, RulesL, SampleSize, Prob, Variance, Head, ClassIndex, ClassT, GeneratedEx),
    append(TrainExList, GeneratedEx, CMMTrainEx),
    tree(CMMTrainEx, BKList, Parameters, Head, TreeKey, ClassIndex, Tree, _DefaultFact),
    evaluate_tree(TestExList, Name, munge, SampleSize, Tree, BKList, Head, TreeKey, ClassIndex, ClassT, PartResult),!,
    use_sampling_methods(R, RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, RestRes).
use_sampling_methods([mbs(SampleSize)|R], RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, [PartResult|RestRes]):-
    write('starting mbs with size: '), write(SampleSize),nl,
    write_to_codes(mbs-TestFold-SampleSize, TreeCodes),
    atom_codes(TreeKey,TreeCodes),
    %set_prolog_flag(profiling, on),
    mbs(RulesL, SampleSize, TrainExList, Head, ClassIndex, ClassT, GeneratedEx),!,
    garbage_collect,
    trimcore,
    %set_prolog_flag(profiling, off),
    append(TrainExList, GeneratedEx, CMMTrainEx),
    tree(CMMTrainEx, BKList, Parameters, Head, TreeKey, ClassIndex, Tree, _DefaultFact),
    evaluate_tree(TestExList, Name, mbs, SampleSize, Tree, BKList, Head, TreeKey, ClassIndex, ClassT, PartResult),!,
    use_sampling_methods(R, RulesL, Name, TrainExList, BKList, TestExList, TestFold, Parameters, Head, ClassIndex, ClassT, RestRes).

evaluate_tree(TestExList, Name, SampMethod, SampleSize, Tree, BKList, Head, TreeKey, ClassIndex, ClassT,
	      result(Name, SampMethod, SampleSize, TreeKey, ClassT, NoRules, TreeEval, TreeEvalTime)):-
      assert_tree_to_bb(Tree, TreeKey, NoRules, _TreeRules),
      statistics(runtime, _),
      classify_test_set(TestExList, ClassIndex, ClassT, ClassT, ClassDist, [TreeKey], TreeRes),
      statistics(runtime, [_,TreeEvalTime]),
      retract_inner_loop(TreeKey, Head, BKList),
      length(TreeRes, TreeResL),
      copy_term(ClassT, CClassT),
      evaluate(ClassDist, CClassT, TreeRes, TreeResL, TreeEval),!.

%Hur hantera NC?! bara lta vxa?! eller mellan regler endast...
remove_all_maj_class_rules([], _, _, _, []).
remove_all_maj_class_rules([rule(_,Conds,_,ClassT)|Rest], true, _, MajClass, NewRevTransRules):-
     get_majority_class(ClassT, MajClass),!,
     remove_all_maj_class_rules(Rest, false, [Conds], MajClass, NewRevTransRules).
remove_all_maj_class_rules([rule(_,Conds,_,ClassT)|Rest], false, PrevNConds, MajClass, NewRevTransRules):-
     get_majority_class(ClassT, MajClass),!,
     remove_all_maj_class_rules(Rest, false, [Conds|PrevNConds], MajClass, NewRevTransRules).
remove_all_maj_class_rules([rule(H,C,NC,CT)|Rest], _, PrevNC, MajClass, [rule(H,C,NNC,CT)|NewRevTransRules]):-
     append(PrevNC,NC,NNC),
     remove_all_maj_class_rules(Rest, true, PrevNC, MajClass, NewRevTransRules).

check_ties(Res, NoTies):-
    check_ties(Res, 0, NoTies).
check_ties([], NoTies, NoTies).
check_ties([_-[No-_,No-_|_]|Rest], TempNoT, NoTies):-
    NewTempNoT is TempNoT + 1,!,
    check_ties(Rest, NewTempNoT, NoTies).
check_ties([_|Rest], TempNoT, NoTies):-
    check_ties(Rest, TempNoT, NoTies).

get_parameters(ensemble(_, Parameters,_), Parameters).
get_parameters(tree(Parameters), Parameters).
get_forrest_size(ensemble(Size, _, _), Size).
get_learning_type(ensemble(_, _, Type), Type).


create_rule_id([], _, []).
create_rule_id([Tree|Trees], ID, FinalRulesWithIDs):-
    create_rule_id1(Tree, ID, NewID, RulesWithIDs1),
    create_rule_id(Trees, NewID, RulesWithIDs2),
    append(RulesWithIDs1, RulesWithIDs2, FinalRulesWithIDs).

create_rule_id1([], NewID, NewID, []).
create_rule_id1([Rule|Rules], ID, FinalID, [ID-Rule|RuleWithIDs]):-
    NewID is ID + 1,
    create_rule_id1(Rules, NewID, FinalID, RuleWithIDs).

evaluate([], _, [], _, []).
evaluate([No-Class|CTuples], ClassTuple, Res, ResL, [Class-[No, TrueP, FalseP, FalseN, FalseTup]|Results]):-
    evaluate_class(Res, Class, RestL, 0, 0, 0, ClassTuple, TrueP, FalseP, FalseN, FalseTup),
    %PercentTrueP is (TrueP / No) * 100,   %ResL
    %PercentFalseP is (FalseP / ResL) * 100, %ResL
    %ClassCov is (No / ResL) * 100 ,
    %nl,
    %write('Class: '), write(Class),nl,
    %write('Covers '), write(ClassCov), write(' of the example space'),nl,
    %write('Accuracy: '), write(PercentTrueP),nl,
    %write('% of False Classifications(wrt. to all classes): '), write(PercentFalseP),nl,
    %write_false_neg1(FalseNTup, ResL, FNegL),!,
    evaluate(CTuples, ClassTuple, RestL, ResL, Results).

write_false_neg1(FalseNTup, TotClassified, Tot-FNegL):-
    sum_false(FalseNTup, 0, Tot),
    %Cov is (Tot / TotClassified) * 100,
    %write('% wrongly classified not belonging to the particular class classified as such are: '), write(Cov),nl,
    write_false_neg(FalseNTup, TotClassified, FNegL).

write_false_neg([], _, []).
write_false_neg(_, 0, 0). %:-
    %write('No False negatives classifications'),nl,!.
write_false_neg([No-Class|Rest], Tot, [Class-Percent|FNegL]):-
    Percent is No/Tot * 100,
    %write('Class '), write(Class), write(' wrongly classified in (%, wrt. all classes): '), write(Percent),nl,
    write_false_neg(Rest, Tot, FNegL).

sum_false([], Tot, Tot).
sum_false([No-_|Rest], TTot, Tot):-
    NewTTot is TTot + No,
    sum_false(Rest, NewTTot, Tot).

evaluate_class([], _Class, [], TrueP, FalseP, FalseN, False, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    NewTTrueP is TTrueP + 1,!,
    evaluate_class(Rest, Class, RestL, NewTTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class|R]|Rest], Class, [Class2-[No-Class|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseP is TFalseP + 1,
    update_false(TFalse, Class, _NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, NewTFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class2|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseN is TFalseN + 1,
    update_false(TFalse, Class2, NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, NewTFalseN, NewTFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class3|R]|Rest], Class, [Class2-[No-Class3|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).

update_false(TFalseN, Class, NewTFalseN):-
    select(No-Class, TFalseN, RestTFalseN),
    NewNo is No + 1,!,
    append([NewNo-Class], RestTFalseN, NewTFalseN).

%prune - write later if needed

classify_test_set([],_ClassIndex, _ClassTuples, FinalCTup, FinalCTup, _Keys, []).
classify_test_set([TestEx|TestExL], ClassIndex, ClassTuples, TempCTup, FCTup, Keys, [PartRes|Res]):-
    TestEx =..[F|Args],
    nth1(ClassIndex, Args, Class, Rest),
    nth1(ClassIndex, NewArgs, _, Rest),
    select(No-Class, TempCTup, TCTup),
    NewNo is No + 1,
    append([NewNo-Class], TCTup, NewTempCTup),
    NTestEx =..[F|NewArgs],
    get_votes(Keys, NTestEx, ClassIndex, Class-ClassTuples, PartRes),!,
    classify_test_set(TestExL, ClassIndex, ClassTuples, NewTempCTup, FCTup, Keys, Res).

get_votes([], _TestEx, _ClassIndex, TClass-ClassTuples, TClass-RSClassT):-
    sort(ClassTuples, SClassT),
    reverse(SClassT, RSClassT).
get_votes([Key|Keys], TestEx, ClassIndex, TClass-ClassTuples, CTuples):-
    copy_term(TestEx,CTestEx),
    call(Key:CTestEx),
    CTestEx =.. [_|Args],
    nth1(ClassIndex, Args, ClassDist),
    keysort(ClassDist,Classes),
    reverse(Classes,[_-Class|_]),
    select(No-Class, ClassTuples, Rest),
    NewNo is No + 1,
    append([NewNo-Class], Rest, NewClassTuples),!,
    get_votes(Keys, TestEx, ClassIndex, TClass-NewClassTuples, CTuples).

do_induction(ensemble(Size, Parameters, dac), Head, TrainExLists, BKList, BBKey, ClassIndex, BBKeys, Trees, NoTrees, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    %create_bag(TrainExList, NewTrainExList),!,
    do_trees(Size, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, BBKeys, Trees, NoTrees, Rules).

do_induction(tree(Parameters), Head, TrainExLists, BKList, BBKey, ClassIndex, BBKeys, Trees, NoTrees, Rules):-
    join_folds(TrainExLists, NewTrainExList),!,
    do_trees(1, Parameters, Head, NewTrainExList, BKList, BBKey, ClassIndex, BBKeys, Trees, NoTrees, Rules).

%%
% join the different folds to one list
join_folds([], []).
join_folds([Fold|Folds], TrainExList):-
   join_folds(Folds, PartList),!,
   append(Fold, PartList, TrainExList).

%%
% random sampling for bagging
create_bag(TrainExList, Size, NewTrainExList):-
    length(TrainExList, NoExamples),!,
    create_bag(TrainExList, Size, NoExamples, NoExamples, NewTrainExList).
create_bag(TrainExList, Size, NewTrainExList):-
    write('Failed creating bag'),nl,
    trace,
    length(TrainExList, NoExamples),
    create_bag(TrainExList, Size, NoExamples, NoExamples, NewTrainExList).
create_bag(_, _, 0, _, []).
create_bag(TrainExList, FSize, NoExLeft, Size, [Example|Bag]):-
    random(0, Size, Rand),
    nth0(Rand, TrainExList, Example),
    NewNoExLeft is NoExLeft - 1,!,
    create_bag(TrainExList, FSize, NewNoExLeft, Size, Bag).
create_bag(TrainExList, FSize, NoExLeft, Size, Bag):-
    write('Failed creating random bag ex'), %try again
    create_bag(TrainExList, FSize, NoExLeft, Size, Bag).

do_trees(0, _Parameters, _Head, _TrainExList, _BKList, _BKKey, _ClassIndex, [], [], [], []).
do_trees(Size, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, [TreeKey|Keys], [Tree|Trees], [NoTrees|Rest], [Rules|RRules]):-
    (member(bagging, Parameters) ->
         %write('Creating bag...'),nl,
         create_bag(TrainExList, Size, NewTrainExList)
         %write('Bag complete...'),nl
    ;
         NewTrainExList = TrainExList
    ),
    write_to_codes(Size-BBKey, TreeCodes),
    atom_codes(TreeKey,TreeCodes),
    %term_to_atom(Size-BBKey, TreeKey),
    %write('Calling tree build'),nl,
    tree(NewTrainExList, BKList, Parameters, Head, TreeKey, ClassIndex, Tree, _DefaultFact),
    %write(Tree), nl,
    %write('Call success'),nl,
    assert_tree_to_bb(Tree, TreeKey, NoTrees, Rules),
    %@todo - create a predicte for only creating rules and not assering them, as we want them to be turned into indexes rewerite to_bb for this cause??!
    %write('Asserted tree to bb'),nl,
    %TreeKey:assertz(DefaultFact),   % default class
    %asserta trden hr med key som bb nyckel? formera dem frst? s de blir en clausul? Eller ha kvar dem i en lista?
    NewSize is Size - 1,!,
    do_trees(NewSize, Parameters, Head, TrainExList, BKList, BBKey, ClassIndex, Keys, Trees, Rest, RRules).

% we only have a default rule, no more rules... lgg till denna i andra kod baser..
assert_tree_to_bb(dac_node(Head,Features,ClassDist,clean_enough), BBKey, 1, [CHead]):-
    copy_term((Head, Features), (CHead, CFeatures)),
    member(class(Var),CFeatures),
    Var = ClassDist,
    BBKey:assertz(CHead).   
assert_tree_to_bb(dac_node(Head,_Features,_ClassDist,Body),BBKey, NoRules, Rules):-
    assert_numerical(BBKey),
    %append(ensemble_rules_home),
    assert_body_to_bb(Head,_,Body,BBKey, 0, NoRules, Rules).
    %write('%One Fold'),nl,
    %told.	
assert_tree_to_bb(dac_node(Head,Features,ClassDist,Body),BBKey, NoRules, Rules):-
     write('ASSERT_TREE_TO_BB FAILED!'),nl,
     write('Head: '),write(Head),nl,
     write('Features: '),write(Features),nl,
     write('ClassDist: '),write(ClassDist),nl,
     write('Body: '),write(Body),nl,
     write('BBKey: '),write(BBKey),nl,
     write('NoRules '),write(NoRules),nl,
     write('Rules '),write(Rules),nl,
     trace,
     assert_body_to_bb(Head,_,Body,BBKey, 0, NoRules, Rules).

%Add more for categorical from Background knowledge etc...
assert_numerical(BBKey):-
    BBKey:asserta(:-(numerical(A),number(A))).

assert_body_to_bb(_Head,_Conds,[],_BBKey, NoRules, NoRules, []).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,no_examples_left)|R],BBKey, TNoRules, NoRules, [:-(CHead,CCond)|Rules]):-
    var(Conds),!, %ignore it
    %write('No examples left 1'),nl,
    copy_term((Head,FList,Cond), (CHead,CFList,CCond)),
    member(class(Var), CFList),
    %keysort(ClassDist,Classes),
    %reverse(Classes,[_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    BBKey:assertz(:-(CHead,CCond)),
    %get_rule_no(RuleNo),
    %portray_clause(:-(CHead,(CCond,write(RuleNo)))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,min_cov)|R],BBKey, TNoRules, NoRules, [:-(CHead,CCond)|Rules]):-
    var(Conds),!, %ignore it
    %write('No examples left 1'),nl,
    copy_term((Head,FList,Cond), (CHead,CFList,CCond)),
    member(class(Var), CFList),
    %keysort(ClassDist,Classes),
    %reverse(Classes,[_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    BBKey:assertz(:-(CHead,CCond)),
    %get_rule_no(RuleNo),
    %portray_clause(:-(CHead,(CCond,write(RuleNo)))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,clean_enough)|R],BBKey, TNoRules, NoRules, [:-(CHead,CCond)|Rules]):-
    var(Conds),!, %ignore it
    copy_term((Head,FList,Cond), (CHead,CFList,CCond)),
    member(class(Var), CFList),
    %keysort(ClassDist,Classes),
    %reverse(Classes,[_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    BBKey:assertz(:-(CHead, CCond)),
    %get_rule_no(RuleNo),
    %BBKey:assertz(:-(CHead, (CCond,write(RuleNo)))),
    %portray_clause(:-(CHead,(CCond,write(RuleNo)))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,no_examples_left)|R],BBKey, TNoRules, NoRules, [:-(CHead,(CCond, CConds))|Rules]):-
    %write('No examples left 2'),nl,
    copy_term((Head, FList, Cond, Conds), (CHead, CFList, CCond, CConds)),
    member(class(Var), CFList),
    %keysort(ClassDist, Classes),
    %reverse(Classes, [_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    %get_rule_no(RuleNo),
    BBKey:assertz(:-(CHead, (CCond, CConds))),
    %BBKey:assertz(:-(CHead,(CCond, (CConds, write(RuleNo))))),
    %portray_clause(:-(CHead,(CCond,(CConds, write(RuleNo))))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,min_cov)|R],BBKey, TNoRules, NoRules, [:-(CHead,(CCond, CConds))|Rules]):-
    %write('No examples left 2'),nl,
    copy_term((Head, FList, Cond, Conds), (CHead, CFList, CCond, CConds)),
    member(class(Var), CFList),
    %keysort(ClassDist, Classes),
    %reverse(Classes, [_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    BBKey:assertz(:-(CHead, (CCond, CConds))),
    %get_rule_no(RuleNo),
    %BBKey:assertz(:-(CHead,(CCond, (CConds, write(RuleNo))))),
    %portray_clause(:-(CHead,(CCond,(CConds, write(RuleNo))))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,FList,ClassDist,clean_enough)|R],BBKey, TNoRules, NoRules, [:-(CHead,(CCond, CConds))|Rules]):-
    copy_term((Head, FList, Cond, Conds), (CHead, CFList, CCond, CConds)),
    member(class(Var), CFList),
    %keysort(ClassDist, Classes),
    %reverse(Classes, [_-Class|_]),
    %Var = Class,
    Var = ClassDist,
    BBKey:assertz(:-(CHead, (CCond, CConds))),
    %get_rule_no(RuleNo),
    %BBKey:assertz(:-(CHead,(CCond, (CConds, write(RuleNo))))),
    %portray_clause(:-(CHead,(CCond,(CConds, write(RuleNo))))),
    %put_rule_no,
    NewTNoRules is TNoRules + 1,
    assert_body_to_bb(Head,Conds,R,BBKey, NewTNoRules, NoRules, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,_FList,_ClassDist,List)|R],BBKey, TNoRules, NoRules, Rules):-
    var(Conds),!,
    assert_body_to_bb(Head,Cond,List,BBKey, 0, NoRules1, Rules1),
    assert_body_to_bb(Head,Conds,R,BBKey, 0, NoRules2, Rules2),
    NoRules is NoRules1 + NoRules2 + TNoRules,
    append(Rules1, Rules2, Rules).
assert_body_to_bb(Head,Conds,[dac_node(Cond,_FList,_ClassDist,List)|R],BBKey, TNoRules, NoRules, Rules):-
    assert_body_to_bb(Head,(Cond,Conds),List,BBKey, 0, NoRules1, Rules1),
    assert_body_to_bb(Head,Conds,R,BBKey, 0, NoRules2, Rules2),
    NoRules is NoRules1 + NoRules2 + TNoRules,
    append(Rules1, Rules2, Rules).

loop_write(_,[]).
loop_write(N,[_F|R]):-
    %length(F,L),
    %write('Fold '),write(N), write(' has '), write(L),write(' examples'),nl,
    NewN is N + 1,
    loop_write(NewN,R).

pre_process_data(N, VSize, ExList, ExLists):- %, ValExList):-
    length(ExList, NoEx),
    ExPerFold is round(NoEx / N),
    NoValPerFold is round(ExPerFold * VSize),
    NewExPerFold is ExPerFold - NoValPerFold,
    rnd_fold_gen(N, ExList, NewExPerFold, NoValPerFold, NoEx, ExLists). %, ValExList).

rnd_fold_gen(0, [], _ExPerFold, _NoValPerFold, _NoEx, []). %, []).
rnd_fold_gen(0, _ExLeft, _, _, _, []). %, []). % throw example away!, fix better solution later
rnd_fold_gen(N, Examples, ExPerFold, NoValsPerFold, NoEx, [Fold|Folds]):- %, [FoldVal|FoldVals]):-
    %gen_one_fold(NoValsPerFold, NoEx, Examples, NewNoEx, NewExamples, FoldVal),
    gen_one_fold(ExPerFold, NoEx, Examples, NewNoEx, NewExamples, Fold),
    NewN is N - 1,
    rnd_fold_gen(NewN, NewExamples, ExPerFold, NoValsPerFold, NewNoEx, Folds). %, FoldVals).

gen_one_fold(0, NoEx, Examples, NoEx, Examples, []).
gen_one_fold(_, 0, Examples, 0, Examples, []).
gen_one_fold(ExPerFold, NoEx, Examples, RestNoEx, RestExamples, [Ex|Exs]):-
    %Pick training examples per fold randomly
    random(0, NoEx, Rand),
    nth0(Rand, Examples, Ex, NewExamples),
    NewExPerFold is ExPerFold - 1,
    NewNoEx is NoEx - 1,
    gen_one_fold(NewExPerFold, NewNoEx, NewExamples, RestNoEx, RestExamples, Exs).

% Read clauses from file
read_from_file(FileName,FileInformation):-
    file_exists(FileName),!,
    see(FileName),
    read(X),
    read_loop(X,FileInformation),
    seen.
read_from_file(_FileName,_FileInformation):-
    write('No such file exist'),nl.

read_loop(end_of_file,[]).
read_loop(X,[X|R]):-
    read(NewX),
    read_loop(NewX,R).

:- setup.



















